import { NestedTreeControl } from '@angular/cdk/tree';
import { AfterViewInit, Component, ElementRef, OnInit, ViewChild } from '@angular/core';
import { FormBuilder, FormGroup } from '@angular/forms';
import { MatDialog } from '@angular/material/dialog';
import { MatTreeNestedDataSource } from '@angular/material/tree';
import { ActivatedRoute, Router } from '@angular/router';
import { map } from 'rxjs/operators';

import { MenuCategory } from '../../core/menu-category';
import { ModifierCategory } from '../../core/modifier-category';
import { Product } from '../../core/product';
import { ToasterService } from '../../core/toaster.service';
import { ConfirmDialogComponent } from '../../shared/confirm-dialog/confirm-dialog.component';
import { ModifierCategoryService } from '../modifier-category.service';
import { NodeItem } from '../node-item';

@Component({
  selector: 'app-role-detail',
  templateUrl: './modifier-category-detail.component.html',
  styleUrls: ['./modifier-category-detail.component.css'],
})
export class ModifierCategoryDetailComponent implements OnInit, AfterViewInit {
  @ViewChild('nameElement', { static: true }) nameElement?: ElementRef;
  form: FormGroup;
  item: ModifierCategory = new ModifierCategory();
  treeControl = new NestedTreeControl<NodeItem>((node: NodeItem) => node.children);
  dataSource = new MatTreeNestedDataSource<NodeItem>();
  products: Map<string, Product> = new Map<string, Product>();
  productsOfMenuCategory: Map<string, Product[]> = new Map<string, Product[]>();

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private fb: FormBuilder,
    private toaster: ToasterService,
    private dialog: MatDialog,
    private ser: ModifierCategoryService,
  ) {
    // Create form
    this.form = this.fb.group({
      name: '',
      minimum: '',
      maximum: '',
      isActive: '',
    });
    this.dataSource.data = [];
  }

  ngOnInit() {
    this.route.data
      .pipe(
        map((value) => {
          const data = value as { item: ModifierCategory };
          const tree: NodeItem[] = data.item.menuCategories.map((x: MenuCategory) => ({
            id: x.id as string,
            name: x.name,
            children: x.products.map((y: Product) => ({ id: y.id as string, name: y.name })),
          }));
          return { item: data.item, tree };
        }),
      )
      .subscribe((data) => {
        this.showItem(data.item, data.tree);
      });
  }

  showItem(item: ModifierCategory, tree: NodeItem[]) {
    this.item = item;
    this.dataSource.data = tree;
    this.form.patchValue({
      name: this.item.name || '',
      minimum: `${this.item.minimum}`,
      maximum: `${this.item.maximum}`,
      isActive: this.item.isActive,
    });
    this.products = new Map();
    this.productsOfMenuCategory = new Map();
    this.item.menuCategories.forEach((mc: MenuCategory) => {
      mc.products.forEach((p: Product) => {
        this.products.set(p.id as string, p);
        if (!this.productsOfMenuCategory.has(mc.id as string)) {
          this.productsOfMenuCategory.set(mc.id as string, []);
        }
        (this.productsOfMenuCategory.get(mc.id as string) as Product[]).push(p);
      });
    });
    // this.setMenuCategories();
  }

  ngAfterViewInit() {
    setTimeout(() => {
      if (this.nameElement !== undefined) {
        this.nameElement.nativeElement.focus();
      }
    }, 0);
  }

  // eslint-disable-next-line @typescript-eslint/no-unused-vars,class-methods-use-this
  isExpanded(mc: MenuCategory) {
    return true;
  }

  save() {
    this.ser.saveOrUpdate(this.getItem()).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigateByUrl('/modifier-categories');
      },
      (error) => {
        this.toaster.show('Error', error);
      },
    );
  }

  delete() {
    this.ser.delete(this.item.id as string).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigateByUrl('/modifier-categories');
      },
      (error) => {
        this.toaster.show('Error', error);
      },
    );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: { title: 'Delete Modifier Category?', content: 'Are you sure? This cannot be undone.' },
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }

  getItem(): ModifierCategory {
    const formModel = this.form.value;
    this.item.name = formModel.name;
    this.item.minimum = +formModel.minimum;
    this.item.maximum = +formModel.maximum;
    this.item.isActive = formModel.isActive;
    return this.item;
  }

  hasChild = (_: number, node: NodeItem) => !!node.children && node.children.length > 0;

  isProductSelected(node: NodeItem) {
    return (this.products.get(node.id) as Product).enabled;
  }

  isMenuCategorySelected(node: NodeItem) {
    return (this.productsOfMenuCategory.get(node.id) as Product[]).reduce(
      (acc: boolean, current: Product) => acc && current.enabled,
      true,
    );
  }

  isMenuCategoryPartiallySelected(node: NodeItem) {
    const total = (this.productsOfMenuCategory.get(node.id) as Product[]).length;
    const ticked = (this.productsOfMenuCategory.get(node.id) as Product[]).reduce(
      (acc, current) => {
        if (current.enabled) {
          // eslint-disable-next-line no-param-reassign
          acc += 1;
        }
        return acc;
      },
      0,
    );
    return ticked > 0 && ticked < total;
  }

  toggleProductSelection(node: NodeItem) {
    (this.products.get(node.id) as Product).enabled = !(this.products.get(node.id) as Product)
      .enabled;
  }

  toggleMenuCategorySelection(node: NodeItem) {
    const sel = !this.isMenuCategorySelected(node);
    (this.productsOfMenuCategory.get(node.id as string) as Product[]).forEach((p) => {
      p.enabled = sel;
    });
    return sel;
  }
}
